#!/usr/bin/env python

"""Console script for methlab."""

import argparse
import ibdpainting as ip
import argparse 
  
def main():
    parser = argparse.ArgumentParser(description='ibdpainting')

    parser.add_argument('-i', '--input',
        help='Path to an HDF5 file containing genotype data for one or more samples to check.  This should be the output of allel.vcf_to_hdf5().'
        )
    parser.add_argument('-n', '--sample_name',
        help ='Sample name for the individual to check. This must be present in the samples in the input file.'
    )
    parser.add_argument('-r', '--reference',
        help="Path to an HDF5 file containing genotype data for a panel of reference individuals to compare the input indivual against. This should be the output of allel.vcf_to_hdf5()."
    )
    parser.add_argument('-w', '--window_size',
        type=int, default=20000,
        help="Integer window size in base pairs."
    )
    parser.add_argument('--expected_match',
        help="Optional list of sample names in the reference panel that are expected to be ancestors of the test individual.",
        nargs = "+", required=False
    )
    parser.add_argument('--outdir',
        help="Directory to save the output."
    )
    parser.add_argument('--keep_ibd_table', 
        help="If set, write an intermediate text file giving genetic distance between the crossed individual and each candidate at each window in the genome. Defaults to False, because these can be quite large.",
        default=False,
        action=argparse.BooleanOptionalAction
    )
    parser.add_argument('--max_to_plot', 
        help="Optional number of the best matching candidates to plot so that the HTML files do not get too large and complicated. Ignored if this is more than the number of samples. Defaults to 20.",
        type=int, default = 10
    )
    parser.add_argument('--interactive',
        help="If set, save the output plot as an interactive HTML plot including information on candidates within the plot.",
        default=True,
        action=argparse.BooleanOptionalAction
        )
    parser.add_argument('--height',
        help="Height in centimetres of the output PNG file.",
        default=675
        )
    parser.add_argument('--width',
        help="Height in centimetres of the output PNG file.",
        default=900)
    args = parser.parse_args()

    # Data frame of IBD at all positions across the genome, and the plot of this
    itable = ip.ibd_table(args.input, args.reference, args.sample_name, args.window_size)
    scores = ip.ibd_scores(itable)
    fig = ip.plot_ibd_table(itable, args.sample_name, args.expected_match, args.max_to_plot)
    
    if args.keep_ibd_table:
        itable.to_csv(args.outdir + "/" + args.sample_name + "_ibd_table.csv", index=False)
    
    scores.to_csv( args.outdir + "/" + args.sample_name + "_ibd_scores.csv", index=False)

    fig.write_image(
        args.outdir + "/" + args.sample_name + "_plot_ibd.png",
        height = args.height, width = args.width
        )
    
    if args.interactive:
        fig.write_html(args.outdir + "/" + args.sample_name + "_plot_ibd.html")
    
        
    

if __name__ == '__main__':
    main()


import numpy as np
import pytest
import ibdpainting as ip

input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'

def test_load_genotype_data_gives_right_output():
    """
    Check that load_genotype_data gives the right answers when it should.
    """
    out = ip.load_genotype_data(
        input = input,
        reference = reference,
        sample_name = 'S2.15.002'
    )
    real_names = np.array(['S2.15.002', '1158', '6024', '6184', '8249'])
    assert all(
        [ x == y for x,y in zip(out.samples, real_names) ]
    )

    assert len(out.chr) == 550
    assert len(out.pos) == 550
    assert out.geno.shape == (550, 5, 2)

def test_load_genotype_data_fails_if_missing_sample():
    """
    Check that load_genotype_data fails if the sample name is not in the input file.
    """
    with pytest.raises(Exception):
        ip.load_genotype_data(
            input = input,
            reference = reference,
            sample_name = 'not_a_real_sample_name'
        )

def test_load_genotype_data_fails_if_contigs_dont_match():
    with pytest.raises(Exception):
        ip.load_genotype_data(
            input = input,
            reference = chr1,
            sample_name = '1158_2'
        )



import ibdpainting as ip


input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
ref_vcf = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'


def test_split_into_windows_functions():
    vcfd = ip.load_genotype_data(
            input = input,
            reference = reference,
            sample_name = 'S2.15.002'
        )
    split_vcfd = vcfd.split_into_windows(1000)
    assert all( split_vcfd['Chr1:0-1000'].pos >= 0 )
    assert all( split_vcfd['Chr1:0-1000'].pos < 1000 )
    assert all(split_vcfd['Chr1:0-1000'].chr == "Chr1")
    assert len(split_vcfd['Chr1:0-1000'].geno.shape) == 3
    # Check you get only one window per chr if window size >> chr length
    assert len(vcfd.split_into_windows(1000000)) == 2

def test_pairwise_distance_works():
    """
    There are four accessions in the reference panel.
    Test each against the whole panel, and check that one of them comes out as
    identical in each case.
    """
    # 1158
    check_1158 = ip.load_genotype_data(
        input = ref_vcf,
        reference = reference,
        sample_name= '1158'
        ).pairwise_distance()

    assert check_1158[0] == 0
    assert all(check_1158[1:] > 0)

    # 6024
    check_6024 = ip.load_genotype_data(input = ref_vcf, reference = reference,
                sample_name= '6024'
        ).pairwise_distance()

    assert check_6024[1] == 0
    assert all(check_6024[[0,2,3]] > 0)

    # 6184
    check_6184 = ip.load_genotype_data(input = ref_vcf, reference = reference,
                sample_name= '6184'
        ).pairwise_distance()

    assert check_6184[2] == 0
    assert all(check_6184[[0,1,3]] > 0)

    # 8249
    check_8249 = ip.load_genotype_data(input = ref_vcf, reference = reference,
                sample_name= '8249'
        ).pairwise_distance()

    assert check_8249[3] == 0
    assert all(check_8249[:2] > 0)
    
    
def test_missing_data_in_geneticDistance():
    """Test that pairwise_distance returns -9 if all loci are NA.
    """
    vcfd = ip.load_genotype_data(
        input = input,
        reference = reference,
        sample_name = 'S2.15.002'
    )   
    vcfd.geno[:,1] = -1
    assert vcfd.pairwise_distance()[0] == -9


import allel
import h5py
import numpy as np
import ibdpainting.geneticDistance

def load_genotype_data(input, reference, sample_name):
    """
    Import and merge test and reference data files.

    Import genotype data for one or more input samples and a panel of reference samples
    to compare to. Subset each so that the markers are really identical. Merge the
    arrays of genotype calls so that the data for the input appear first on the
    first axis of the genotype call arrays.

    Parameters
    ==========
    input: str
        Path to a an HDF5 file containing genotype data for one or more samples to check
    reference: str
        Path to a HDF5 file containing genotype data for a panel of reference individuals
        to compare the input indivual against.
    sample_name: str
        Sample name for the individual to check. This must be present in the samples
        in the input file.

    Return
    ======
    An object of class geneticDistance.
    """
    # Read in the data files
    input_hdf5 = h5py.File(input, mode='r')
    ref_hdf5  = h5py.File(reference, mode="r")

    input_str_data = {
        'samples' : [ x.decode('utf-8') for x in input_hdf5['samples'][:] ],
        'chr'     : [ x.decode('utf-8') for x in input_hdf5['variants/CHROM'][:] ]
    }

    ref_str_data = {
        'samples' : [ x.decode('utf-8') for x in ref_hdf5['samples'][:] ],
        'chr'     : [ x.decode('utf-8') for x in ref_hdf5['variants/CHROM'][:] ]
    }
    import time

    if sample_name not in input_str_data['samples']:
        raise ValueError("The sample name is not in the list of samples in the input file.")
    else: 
        # Find the position of the individual to test
        sample_ix = np.where(
            [ sample_name == x for x in input_str_data['samples'] ]
            )[0][0]
        # Join vectors of sample names, with the test individual first
        new_samples = np.append(
            input_str_data['samples'][sample_ix], ref_str_data['samples']
            )        

    # Check that contig labels match
    chr_labels = {
        'input' : np.unique(input_str_data['chr']),
        'ref'   : np.unique(ref_str_data['chr'])
    }
    if len(chr_labels['input']) != len(chr_labels['ref']):
        raise ValueError(
            "The number of unique contig labels do not match: the input an HDF5 has {}, but the reference panel has {}.".
            format( chr_labels['input'], chr_labels['ref'] )
        )
    elif any( chr_labels['input'] != chr_labels['ref'] ):
        raise ValueError(
            "Contig labels do not match between the input and reference files."
        )
    
    # Make sure we only compare SNPs that are found in both datasets.
    # Concatenate chromosome labels and SNP positions
    snp_names = {
        'input' : [ str(chr) + ":" + str(pos) for chr,pos in zip(input_str_data['chr'], input_hdf5['variants/POS'][:]) ],
        'ref'   : [ str(chr) + ":" + str(pos) for chr,pos in zip(ref_str_data['chr'], ref_hdf5['variants/POS'][:]) ]
    }
    # Find the SNP position names that are common to both datasets
    matching_SNPs_in_both_files = np.intersect1d(
        snp_names['input'],
        snp_names['ref']
        )
    which_SNPs_to_keep = {
        "input" : [ x in matching_SNPs_in_both_files for x in snp_names['input'] ],
        "ref"   : [ x in matching_SNPs_in_both_files for x in snp_names['ref'] ]
    }


    # Append the genotype data for the test individual to the array of the reference panel
    new_geno = np.concatenate(
        (input_hdf5['calldata/GT'][which_SNPs_to_keep['input'], sample_ix][:, np.newaxis],
        ref_hdf5['calldata/GT'][which_SNPs_to_keep['ref']]),
        axis=1
        )
    
    # Define an output before closing the Hdf5 file
    output = ibdpainting.geneticDistance(
        samples = new_samples,
        chr = np.array(ref_str_data['chr'])[np.where(which_SNPs_to_keep['ref'])[0]],
        pos = ref_hdf5['variants/POS'][:][which_SNPs_to_keep['ref']],
        geno = new_geno
    )

    ref_hdf5.close()

    return output


# Not run - need to work out how to test a CLI properly.

import ibdpainting as ip


input = 'tests/test_data/panel_to_test.hdf5'
reference = 'tests/test_data/reference_panel.hdf5'
ref_vcf = 'tests/test_data/reference_panel.hdf5'
chr1 = 'tests/test_data/reference_panel_chr1.hdf5'


"""
ibdpainting \
    --input tests/test_data/reference_panel.hdf5 \
    --reference tests/test_data/reference_panel.hdf5 \
    --sample_name 1158 \
    --window_size 1000 \
    --outdir tests/test_output \
    --expected_match 1158 \
    --no-interactive
"""


from setuptools import setup
import codecs
import os.path
from pathlib import Path

# Functions to pull the package version from init.py
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='ibdpainting',
    version=get_version("ibdpainting/__init__.py"),
    description='Identify parents of a crossed individual by comparing identity in windows across their genomes',
    url='https://github.com/ellisztamas/ibdpainting',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tom Ellis',
    author_email='thomas.ellis@gmi.oeaw.ac.at',
    license='MIT',
    packages=['ibdpainting'],
    install_requires=[
        'numpy', 'pandas', 'plotly', 'h5py', 'scikit-allel', 'pytest', 'kaleido'
      ],
    zip_safe=False,
    entry_points = {
        'console_scripts': ['ibdpainting=ibdpainting.command_line:main'],
    }
    )




import pandas as pd

from ibdpainting import load_genotype_data

def ibd_table(input:str, reference:str, sample_name:str, window_size:int):
    """
    Compare allele sharing across the genome.

    Calculate genetic distance between a test individual and a panel of
    reference genomes.

    Parameters
    ==========
    input: str
        Path to an HDF5 file containing genotype data for one or more samples to
        test
    reference: str
        Path to an HDF5 file containing genotype data for a panel of reference
        individuals to compare the test individual against.
    sample_name: str
        Sample name for the individual to check.
        This must be present in the samples in the input HDF5 file.
    window_size: int
        Window size in base pairs.

    Returns
    =======
    DataFrame with a row for each window in the genome and a column for each 
    sample in the reference panel. Elements show genetic distance between the 
    test individual and each reference individual in a single window.
    """
    genetic_distance = load_genotype_data(
            input = input,
            reference = reference,
            sample_name = sample_name
        )
    # Divide the genome into windows
    distances_in_windows = genetic_distance.split_into_windows(window_size)

    # Dataframe with a row for each window across the genome and a column for each sample in the reference panel.
    distance_array = pd.DataFrame(
        [ v.pairwise_distance() for v in distances_in_windows.values() ],
        columns = genetic_distance.samples[1:]
    )
    distance_array.insert(0, 'window', distances_in_windows.keys())

    return distance_array


"""Top-level package for ibdpainting."""

__author__ = """Tom Ellis"""
__email__ = 'thomas.ellis@gmi.oeaw.ac.at'
__version__ = '0.2.0'

from ibdpainting.load_genotype_data import load_genotype_data
from ibdpainting.geneticDistance import *
from ibdpainting.ibd_table import ibd_table
from ibdpainting.ibd_scores import ibd_scores
from ibdpainting.plot_ibd_table import plot_ibd_table


import numpy as np
from warnings import warn
import numpy.ma as ma

class geneticDistance(object):
    """
    A simple class to compare genotype data genetic distances between individuals.  

    Parameters
    ==========
    samples: array
        Vector of length m giving names for each sample.
    chr: array
        Vector of length n giving chromosome labels for each SNP.
    pos: array
        Vector of length n giving SNP positions. Note that SNP positions are inherited from 
        skikit allel and give row numbers from the input HDF5 file rather than
        base-pair positions on each chromosome.
    geno: array
        m x n x 2 array of genotype data where axes index SNPs, individuals, and 
        homologous chromosomes.

    Attributes
    ==========
    samples: array
        Vector of M sample names. The first sample is the input individual to be
        compared to the remaining reference individuals.
    chr: array
        Vector of chromosome labels. These are imported from the reference panel.
    pos: array
        Vector of N SNP positions. These are imported from the reference panel.
    geno: array
        NxMx2 array of genotype data, where N is the number of SNPs and M is the
        number of samples.

    Methods
    =======
    split_into_windows
        Split a geneticDistance object into windows.
    pairwise_distance
        Calculate pairwise genetic distance between an input individual and all 
        reference individuals.
    
    """
    def __init__(self, samples, chr, pos, geno):
        self.samples = samples
        self.chr = chr
        self.pos = pos
        self.geno = geno

    def split_into_windows(self, window_size: int):
        """
        Split a geneticDistance object into windows.

        Splits the geneticDistance object into chromosomes, then into windows on each
        chromosome. It returns a dictionary of geneticDistance objects for each window.

        Parameters
        ==========
        window_size: int
            Window size in base pairs.

        Returns
        =======
        A dictionary of geneticDistance objects with an element for each window.
        Indexes are in the form "Chr:start-stop".
        """
        # Empty dict to store the output
        list_of_distance_objects = {}

        for chr in np.unique(self.chr):
            # Boolean array indexing items in this chromosome
            chr_ix = self.chr == chr
            
            # Array of starting positions for each window. 
            start_positions = np.arange(0, self.pos[chr_ix].max(), window_size)
            for start in start_positions:
                stop  = start + window_size
                # Index positions of SNPs within the current window
                window_ix = (self.pos[chr_ix] >= start) & (self.pos[chr_ix] < stop)
                # Create an object for each window.
                window_name = str(chr) + ":" + str(start) + "-" + str(stop)
                list_of_distance_objects[window_name] = geneticDistance(
                        samples = self.samples,
                        chr  = self.chr[chr_ix][window_ix],
                        pos  = self.pos[chr_ix][window_ix],
                        geno = self.geno[chr_ix][window_ix]
                    )
        
        return list_of_distance_objects
    
    def pairwise_distance(self):
        """
        Calculate pairwise genetic distance between an input individual and all 
        reference individuals.

        The input individual is always the first in the list of samples. Genetic
        distance is the number of allelic differences at each locus between each
        pair, summed over all loci. The calculation is done using masked arrays to
        account for missing data.

        Returns
        =======
        Vector of distances

        """
        masked_geno = ma.masked_array(self.geno, self.geno < 0)

        # Calculate differences at each locus
        per_locus_difference = abs(masked_geno.sum(2)[:,[0]] - masked_geno.sum(2)[:,1:]) / 2
        # Average over loci
        dxy = per_locus_difference.mean(0)
        
        return ma.filled(dxy, -9)
